///////////////////////////////////////////////////////////////////
// Functy
// 3D graph drawing utility
//
// David Llewellyn-Jones
// http://www.flypig.co.uk
//
// Summer 2012
///////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////
// Includes

#include "exportply.h"

///////////////////////////////////////////////////////////////////
// Defines

#ifdef WORDS_BIGENDIAN
#define PLY_ENDIANNESS "binary_big_endian"
#else
#define PLY_ENDIANNESS "binary_little_endian"
#endif

///////////////////////////////////////////////////////////////////
// Structures and enumerations

///////////////////////////////////////////////////////////////////
// Global variables

///////////////////////////////////////////////////////////////////
// Function prototypes

///////////////////////////////////////////////////////////////////
// Function definitions

bool ExportModel (char const * szFilename, bool boBinary, bool boScreenCoords, bool boExportAlpha, double fMultiplier, GSList const * const psFunctions) {
	bool boResult;
	int nVertices;
	int nFaces;
	FILE * hFile;
	int nOffset;
  GSList const * psFuncList;
  GSList * psOffsets;
  GSList * psOffsetCurrent;

	// Calculate the number of vertices
	nVertices = 0;
  psFuncList = psFunctions;
  while (psFuncList) {
		nVertices += CountStoredVertices (fMultiplier, (FuncPersist const *)(psFuncList->data));
    psFuncList = g_slist_next (psFuncList);
  }

	// Calculate the number of faces
	nFaces = 0;
  psFuncList = psFunctions;
  while (psFuncList) {
		nFaces += CountStoredFaces (fMultiplier, (FuncPersist const *)(psFuncList->data));
    psFuncList = g_slist_next (psFuncList);
  }

	boResult = FALSE;
	hFile = fopen (szFilename, "w");
	if (hFile) {
		fprintf (hFile, "ply\n");
		if (boBinary) {
			fprintf (hFile, "format " PLY_ENDIANNESS " 1.0 1.0\n");
		}
		else {
			fprintf (hFile, "format ascii 1.0\n");
		}
		fprintf (hFile, "comment Functy generated\n");
		fprintf (hFile, "element vertex %d\n", nVertices);
		fprintf (hFile, "property float32 x\n");
		fprintf (hFile, "property float32 y\n");
		fprintf (hFile, "property float32 z\n");
		fprintf (hFile, "property uint8 red\n");
		fprintf (hFile, "property uint8 green\n");
		fprintf (hFile, "property uint8 blue\n");
		if (boExportAlpha) {
			fprintf (hFile, "property uint8 alpha\n");
		}
		fprintf (hFile, "property float32 nx\n");
		fprintf (hFile, "property float32 ny\n");
		fprintf (hFile, "property float32 nz\n");
		fprintf (hFile, "element face %d\n", nFaces);
		fprintf (hFile, "property list uchar int vertex_indices\n");
		fprintf (hFile, "end_header\n");

		// Output the vertices
	  psOffsets = NULL;
		psFuncList = psFunctions;
		while (psFuncList) {
			nVertices = OutputStoredVertices (hFile, boBinary, boScreenCoords, boExportAlpha, fMultiplier, (FuncPersist const *)(psFuncList->data));
		  psOffsets = g_slist_append (psOffsets, GINT_TO_POINTER (nVertices));
		  psFuncList = g_slist_next (psFuncList);
		}
		
		// Output the indices
		nOffset = 0;
		psOffsetCurrent = psOffsets;
		psFuncList = psFunctions;
		while (psFuncList) {
			OutputStoredIndices (hFile, boBinary, fMultiplier, nOffset, (FuncPersist const *)(psFuncList->data));
			nOffset += GPOINTER_TO_INT (psOffsetCurrent->data);
			psOffsetCurrent = g_slist_next (psOffsetCurrent);
		  psFuncList = g_slist_next (psFuncList);
		}
		g_slist_free (psOffsets);

		fclose (hFile);
		boResult = TRUE;
	}

	return boResult;
}


