///////////////////////////////////////////////////////////////////
// Functy
// 3D graph drawing utility
//
// David Llewellyn-Jones
// http://www.flypig.co.uk
//
// Spring 2009
///////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////
// Includes

#include "vis.h"
#include "function.h"
#include "function_private.h"
#include "textures.h"

///////////////////////////////////////////////////////////////////
// Defines

#define FUNCTION0 "1"

///////////////////////////////////////////////////////////////////
// Structures and enumerations

///////////////////////////////////////////////////////////////////
// Global variables

///////////////////////////////////////////////////////////////////
// Function prototypes

void GenerateVertices (FuncPersist * psFuncData);

///////////////////////////////////////////////////////////////////
// Function definitions

FuncPersist * NewFuncPersist (FUNCTYPE eType) {
  FuncPersist * psFuncData;

  psFuncData = g_new0 (FuncPersist, 1);

  psFuncData->eType = eType;
  switch (eType) {
  case FUNCTYPE_CARTESIAN:
    psFuncData->Func.psCartesianData = NewCartesianPersist ();
    break;
  case FUNCTYPE_SPHERICAL:
    psFuncData->Func.psSphericalData = NewSphericalPersist ();
    break;
  case FUNCTYPE_CURVE:
  	psFuncData->Func.psCurveData = NewCurvePersist ();
  	break;
  default:
    psFuncData->Func.psNone = NULL;
    break;
  }

  psFuncData->fXMin = -5.0;
  psFuncData->fYMin = -5.0;
  psFuncData->fZMin = -20.0;
  psFuncData->fXWidth = 10.0;
  psFuncData->fYWidth = 10.0;
  psFuncData->fZWidth = 40.0;

  psFuncData->fAccuracy = ACCURACY_NEW;
  psFuncData->afVertices = NULL;
  psFuncData->afNormals = NULL;
  psFuncData->auIndices = NULL;
  psFuncData->afColours = NULL;
  psFuncData->afTextureCoords = NULL;

  psFuncData->psFunction = NULL;
  psFuncData->psVariables = NULL;

  psFuncData->boColourFunction = FALSE;
  psFuncData->fRed = 0.5f;
  psFuncData->fGreen = 0.5f;
  psFuncData->fBlue = 0.5f;
  psFuncData->fAlpha = 0.5f;
  psFuncData->psRed = NULL;
  psFuncData->psGreen = NULL;
  psFuncData->psBlue = NULL;
  psFuncData->psAlpha = NULL;

	psFuncData->uTexture = TEXTURE_NONE;
  psFuncData->szTexFile = g_string_new ("");
  psFuncData->szTexXScale = g_string_new ("1.0");
  psFuncData->szTexYScale = g_string_new ("1.0");
  psFuncData->szTexXOffset = g_string_new ("0.0");
  psFuncData->szTexYOffset = g_string_new ("0.0");

  psFuncData->szFunction = g_string_new ("");
  psFuncData->szRed = g_string_new ("1.0");
  psFuncData->szGreen = g_string_new ("0.5");
  psFuncData->szBlue = g_string_new ("0.5");
  psFuncData->szAlpha = g_string_new ("0.9");

  psFuncData->boTimeDependent = FALSE;
  psFuncData->psShaderData = NULL;
  psFuncData->szShaderVertexSource = NULL;
  psFuncData->szShaderFragmentSource = NULL;

  psFuncData->szShaderFunction = g_string_new ("0");
  psFuncData->szShaderRed = g_string_new ("1");
  psFuncData->szShaderGreen = g_string_new ("1");
  psFuncData->szShaderBlue = g_string_new ("1");
  psFuncData->szShaderAlpha = g_string_new ("1");

	InitShader (psFuncData);

  SetFunction (FUNCTION0, psFuncData);

  GenerateVertices (psFuncData);
  PopulateVertices (psFuncData);

  return psFuncData;
}

FUNCTYPE GetFunctionType (FuncPersist * psFuncData) {
  return psFuncData->eType;
}

void DeleteFuncPersist (FuncPersist * psFuncData) {
  // Free up the subtype
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    DeleteCartesianPersist (psFuncData->Func.psCartesianData);
    psFuncData->Func.psCartesianData = NULL;
    break;
  case FUNCTYPE_SPHERICAL:
    DeleteSphericalPersist (psFuncData->Func.psSphericalData);
    psFuncData->Func.psSphericalData = NULL;
    break;
  case FUNCTYPE_CURVE:
    DeleteCurvePersist (psFuncData->Func.psCurveData);
    psFuncData->Func.psCurveData = NULL;
  	break;
  default:
    // Do absolutely nothing
    break;
  }

  // Free up the function data
  FreeRecursive (psFuncData->psFunction);

  // Free up the colour data
  FreeRecursive (psFuncData->psRed);
  FreeRecursive (psFuncData->psGreen);
  FreeRecursive (psFuncData->psBlue);

  psFuncData->psVariables = FreeVariables (psFuncData->psVariables);

  // Free up the handy UI-related data
  if (psFuncData->szFunction) {
    g_string_free (psFuncData->szFunction, TRUE);
  }
  if (psFuncData->szRed) {
    g_string_free (psFuncData->szRed, TRUE);
  }
  if (psFuncData->szGreen) {
    g_string_free (psFuncData->szGreen, TRUE);
  }
  if (psFuncData->szBlue) {
    g_string_free (psFuncData->szBlue, TRUE);
  }
  if (psFuncData->szAlpha) {
    g_string_free (psFuncData->szAlpha, TRUE);
  }

  if (psFuncData->szTexFile) {
    g_string_free (psFuncData->szTexFile, TRUE);
  }
  if (psFuncData->szTexXScale) {
    g_string_free (psFuncData->szTexXScale, TRUE);
  }
  if (psFuncData->szTexYScale) {
    g_string_free (psFuncData->szTexYScale, TRUE);
  }
  if (psFuncData->szTexXOffset) {
    g_string_free (psFuncData->szTexXOffset, TRUE);
  }
  if (psFuncData->szTexYOffset) {
    g_string_free (psFuncData->szTexYOffset, TRUE);
  }

  // Free up the OpenGL vertex and normal datas
	FreeVertexBuffers (psFuncData);
  
  // Free up any texture
  if (psFuncData->uTexture != TEXTURE_NONE) {
  	UnloadTexture (psFuncData->uTexture);
  }

	// Free the shader data
	if (psFuncData->psShaderData) {
		DeleteShaderPersist (psFuncData->psShaderData);
		psFuncData->psShaderData = NULL;
	}

	if (psFuncData->szShaderVertexSource) {
		g_free (psFuncData->szShaderVertexSource);
		psFuncData->szShaderVertexSource = NULL;
	}

	if (psFuncData->szShaderFragmentSource) {
		g_free (psFuncData->szShaderFragmentSource);
		psFuncData->szShaderFragmentSource = NULL;
	}

  if (psFuncData->szShaderFunction) {
    g_string_free (psFuncData->szShaderFunction, TRUE);
  }
  if (psFuncData->szShaderRed) {
    g_string_free (psFuncData->szShaderRed, TRUE);
  }
  if (psFuncData->szShaderGreen) {
    g_string_free (psFuncData->szShaderGreen, TRUE);
  }
  if (psFuncData->szShaderBlue) {
    g_string_free (psFuncData->szShaderBlue, TRUE);
  }
  if (psFuncData->szShaderAlpha) {
    g_string_free (psFuncData->szShaderAlpha, TRUE);
  }

  g_free (psFuncData);
}

void InitShader (FuncPersist * psFuncData) {
	if (psFuncData->psShaderData == NULL) {
		psFuncData->psShaderData = NewShaderPersist ();
	}

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
		CartesianInitShader (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
		SphericalInitShader (psFuncData);
    break;
  case FUNCTYPE_CURVE:
		CurveInitShader (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void LoadVertexShader (char const * const szFilename, FuncPersist * psFuncData) {
	// Free up any previously loaded shader source
	if (psFuncData->szShaderVertexSource) {
		g_free (psFuncData->szShaderVertexSource);
		psFuncData->szShaderVertexSource = NULL;
	}

	psFuncData->szShaderVertexSource = LoadShaderFile (szFilename);
}

void LoadFragmentShader (char const * const szFilename, FuncPersist * psFuncData) {
	// Free up any previously loaded shader source
	if (psFuncData->szShaderFragmentSource) {
		g_free (psFuncData->szShaderFragmentSource);
		psFuncData->szShaderFragmentSource = NULL;
	}

	psFuncData->szShaderFragmentSource = LoadShaderFile (szFilename);
}

void ActivateFunctionShader (FuncPersist * psFuncData) {
	ActivateShader (psFuncData->psShaderData);
}

void DeactivateFunctionShader (FuncPersist * psFuncData) {
	DeactivateShader (psFuncData->psShaderData);
}

void SetFunctionShaderActive (bool boActive, FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
  	CartesianSetShaderActive (boActive, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
  	SphericalSetShaderActive (boActive, psFuncData);
    break;
  case FUNCTYPE_CURVE:
  	CurveSetShaderActive (boActive, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void SetFunction (char const * const szFunction, FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianSetFunction (szFunction, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalSetFunction (szFunction, psFuncData);
    break;
  case FUNCTYPE_CURVE:
		CurveSetFunction ("a", "0", "0", szFunction, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void SetFunctionColours (char const * const szRed, char const * const szGreen, char const * const szBlue, char const * const szAlpha, FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianSetFunctionColours (szRed, szGreen, szBlue, szAlpha, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalSetFunctionColours (szRed, szGreen, szBlue, szAlpha, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveSetFunctionColours (szRed, szGreen, szBlue, szAlpha, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void FunctionShadersRegenerate (FuncPersist * psFuncData) {
	char * szShader;

	szShader = NULL;
	if (psFuncData->psShaderData) {
		// Generate a vertex shader
		switch (psFuncData->eType) {
		case FUNCTYPE_CARTESIAN:
			szShader = CartesianGenerateVertexShader (psFuncData);
		  break;
		case FUNCTYPE_SPHERICAL:
			szShader = SphericalGenerateVertexShader (psFuncData);
		  break;
		case FUNCTYPE_CURVE:
			szShader = CurveGenerateVertexShader (psFuncData);
			break;
		default:
		  // Do absolutely nothing
		  break;
		}

		ShaderRegenerateVertex (szShader, psFuncData->psShaderData);
		if (szShader) {
			g_free (szShader);
			szShader = NULL;
		}

		// Generate a fragment shader
		switch (psFuncData->eType) {
		case FUNCTYPE_CARTESIAN:
			szShader = CartesianGenerateFragmentShader (psFuncData);
		  break;
		case FUNCTYPE_SPHERICAL:
			szShader = SphericalGenerateFragmentShader (psFuncData);
		  break;
		case FUNCTYPE_CURVE:
			szShader = CurveGenerateFragmentShader (psFuncData);
			break;
		default:
		  // Do absolutely nothing
		  break;
		}

		ShaderRegenerateFragment (szShader, psFuncData->psShaderData);
		if (szShader) {
			g_free (szShader);
			szShader = NULL;
		}

		printf ("Shaders regenerated.\n");
	}
}

char const * GetFunctionString (FuncPersist * psFuncData) {
  return psFuncData->szFunction->str;
}

char const * GetRedString (FuncPersist * psFuncData) {
  return psFuncData->szRed->str;
}

char const * GetGreenString (FuncPersist * psFuncData) {
  return psFuncData->szGreen->str;
}

char const * GetBlueString (FuncPersist * psFuncData) {
  return psFuncData->szBlue->str;
}

char const * GetAlphaString (FuncPersist * psFuncData) {
  return psFuncData->szAlpha->str;
}

char const * GetTexFileString (FuncPersist * psFuncData) {
  return psFuncData->szTexFile->str;
}

char const * GetTexXOffsetString (FuncPersist * psFuncData) {
  return psFuncData->szTexXOffset->str;
}

char const * GetTexYOffsetString (FuncPersist * psFuncData) {
  return psFuncData->szTexYOffset->str;
}

char const * GetTexXScaleString (FuncPersist * psFuncData) {
  return psFuncData->szTexXScale->str;
}

char const * GetTexYScaleString (FuncPersist * psFuncData) {
  return psFuncData->szTexYScale->str;
}

void SetFunctionRange (double fXMin, double fYMin, double fZMin, double fXWidth, double fYWidth, double fZWidth, FuncPersist * psFuncData) {
	Vector3 vScale;

  psFuncData->fXWidth = fXWidth;
  psFuncData->fYWidth = fYWidth;
  psFuncData->fZWidth = fZWidth;

	vScale.fX = fXWidth / AXIS_XSIZE;
	vScale.fY = fYWidth / AXIS_YSIZE;
	vScale.fZ = fZWidth / AXIS_ZSIZE;
	SetShaderScale (& vScale, psFuncData->psShaderData);

	SetFunctionPosition (fXMin, fYMin, fZMin, psFuncData);
}

void GetFunctionRange (double * afRange, FuncPersist const * psFuncData) {
  if (afRange) {
    afRange[0] = psFuncData->fXMin;
    afRange[1] = psFuncData->fYMin;
    afRange[2] = psFuncData->fZMin;
    afRange[3] = psFuncData->fXWidth;
    afRange[4] = psFuncData->fYWidth;
    afRange[5] = psFuncData->fZWidth;
  }
}

void SetFunctionPosition (double fXMin, double fYMin, double fZMin, FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianSetFunctionPosition (fXMin, fYMin, fZMin, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalSetFunctionPosition (fXMin, fYMin, fZMin, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveSetFunctionPosition (fXMin, fYMin, fZMin, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void FreeVertexBuffers (FuncPersist * psFuncData) {
  // Free vertex buffer
  if (psFuncData->afVertices) {
    g_free (psFuncData->afVertices);
    psFuncData->afVertices = NULL;
  }
  // Free normal buffer
  if (psFuncData->afNormals) {
    g_free (psFuncData->afNormals);
    psFuncData->afNormals = NULL;
  }
  // Free index buffer
  if (psFuncData->auIndices) {
    g_free (psFuncData->auIndices);
    psFuncData->auIndices = NULL;
  }
  // Free colour buffer
  if (psFuncData->afColours) {
    g_free (psFuncData->afColours);
    psFuncData->afColours = NULL;
  }
  // Free texture coordinate buffer
  if (psFuncData->afTextureCoords) {
    g_free (psFuncData->afTextureCoords);
    psFuncData->afTextureCoords = NULL;
  }
}

void SetFunctionAccuracy (double fAccuracy, FuncPersist * psFuncData) {
  // Don't do anything if the accuracy hasn't changed
  if (psFuncData->fAccuracy != fAccuracy) {
    psFuncData->fAccuracy = fAccuracy;

		FreeVertexBuffers (psFuncData);
    GenerateVertices (psFuncData);
    PopulateVertices (psFuncData);
  }
}

double GetFunctionAccuracy (FuncPersist * psFuncData) {
  return psFuncData->fAccuracy;
}

GLfloat * GetVertices (FuncPersist * psFuncData) {
  return psFuncData->afVertices;
}

GLfloat * GetNormals (FuncPersist * psFuncData) {
  return psFuncData->afNormals;
}

GLushort * GetIndices (FuncPersist * psFuncData) {
  return psFuncData->auIndices;
}

GLfloat * GetColours (FuncPersist * psFuncData) {
  return psFuncData->afColours;
}

GLfloat * GetTextureCoords (FuncPersist * psFuncData) {
  return psFuncData->afTextureCoords;
}

bool GetColour (float * afGraphColour, FuncPersist * psFuncData) {
  if (!psFuncData->boColourFunction) {
    afGraphColour[0] = psFuncData->fRed;
    afGraphColour[1] = psFuncData->fGreen;
    afGraphColour[2] = psFuncData->fBlue;
    afGraphColour[3] = psFuncData->fAlpha;
  }

  return !psFuncData->boColourFunction;
}

void GetVertexDimensions (int * pnV1Vertices, int * pnV2Vertices, FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianGetVertexDimensions (pnV1Vertices, pnV2Vertices, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalGetVertexDimensions (pnV1Vertices, pnV2Vertices, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveGetVertexDimensions (pnV1Vertices, pnV2Vertices, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void PopulateVertices (FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianPopulateVertices (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalPopulateVertices (psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurvePopulateVertices (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void GenerateVertices (FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianGenerateVertices (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalGenerateVertices (psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveGenerateVertices (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

bool GetTimeDependent (FuncPersist * psFuncData) {
  return psFuncData->boTimeDependent;
}

void SetFunctionTime (double fTime, FuncPersist * psFuncData) {
  if (psFuncData->psVariableT) {
    SetVariable (psFuncData->psVariableT, fTime);
  }

	SetShaderTime (fTime, psFuncData->psShaderData);
}

void SetFunctionTexture (GLuint uTexture, FuncPersist * psFuncData) {
	psFuncData->uTexture = uTexture;
}

GLuint GetFunctionTexture (FuncPersist * psFuncData) {
	return psFuncData->uTexture;
}

void LoadTexture (char const * const szFilename, FuncPersist * psFuncData) {
  // Free up any texture
  if (psFuncData->uTexture != TEXTURE_NONE) {
  	UnloadTexture (psFuncData->uTexture);
  	psFuncData->uTexture = TEXTURE_NONE;
  }

	psFuncData->uTexture = LoadTextureRaw (szFilename, 2048, 2048, true);
}

void SetTextureValues (char const * const szFilename, char const * const szXScale, char const * const szYScale, char const * const szXOffset, char const * const szYOffset, FuncPersist * psFuncData) {
	// Load a new texture file if the filename has changed
	if (g_strcmp0 (szFilename, psFuncData->szTexFile->str) != 0) {
	  g_string_assign (psFuncData->szTexFile, szFilename);
		if (psFuncData->uTexture != TEXTURE_NONE) {
			UnloadTexture (psFuncData->uTexture);
			psFuncData->uTexture = TEXTURE_NONE;
		}
		if (g_strcmp0 (psFuncData->szTexFile->str, "") != 0) {
			LoadTexture (szFilename, psFuncData);
		}
	}

	g_string_assign (psFuncData->szTexXScale, szXScale);
	g_string_assign (psFuncData->szTexYScale, szYScale);
	g_string_assign (psFuncData->szTexXOffset, szXOffset);
	g_string_assign (psFuncData->szTexYOffset, szYOffset);
}

char const * GetXCentreString (FuncPersist * psFuncData) {
	char const * szResult = NULL;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    szResult = CartesianGetXCentreString (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    szResult = SphericalGetXCentreString (psFuncData);
    break;
  case FUNCTYPE_CURVE:
    szResult = CurveGetXCentreString (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }

  return szResult;
}

char const * GetYCentreString (FuncPersist * psFuncData) {
	char const * szResult = NULL;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    szResult = CartesianGetYCentreString (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    szResult = SphericalGetYCentreString (psFuncData);
    break;
  case FUNCTYPE_CURVE:
    szResult = CurveGetYCentreString (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }

  return szResult;
}

char const * GetZCentreString (FuncPersist * psFuncData) {
	char const * szResult = NULL;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    szResult = CartesianGetZCentreString (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    szResult = SphericalGetZCentreString (psFuncData);
    break;
  case FUNCTYPE_CURVE:
    szResult = CurveGetZCentreString (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }

  return szResult;
}

void GetCentre (double * afCentre, FuncPersist const * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianGetCentre (afCentre, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalGetCentre (afCentre, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveGetCentre (afCentre, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void UpdateCentre (FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianUpdateCentre (psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalUpdateCentre (psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveUpdateCentre (psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

void SetFunctionCentre (char const * const szXCentre, char const * const szYCentre, char const * const szZCentre, FuncPersist * psFuncData) {
  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    CartesianSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    SphericalSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    CurveSetFunctionCentre (szXCentre, szYCentre, szZCentre, psFuncData);
  	break;
  default:
    // Do absolutely nothing
    break;
  }
}

int OutputStoredVertices (FILE * hFile, bool boBinary, bool boScreenCoords, bool boExportAlpha, double fMultiplier, FuncPersist const * psFuncData) {
	int nVertices;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    nVertices = CartesianOutputStoredVertices (hFile, boBinary, boScreenCoords, boExportAlpha, fMultiplier, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    nVertices = SphericalOutputStoredVertices (hFile, boBinary, boScreenCoords, boExportAlpha, fMultiplier, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    nVertices = CurveOutputStoredVertices (hFile, boBinary, boScreenCoords, boExportAlpha, fMultiplier, psFuncData);
  	break;
  default:
  	// No vertices
    nVertices = 0;
    break;
  }

	return nVertices;
}

int OutputStoredIndices (FILE * hFile, bool boBinary, double fMultiplier, int nOffset, FuncPersist const * psFuncData) {
	int nIndices;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    nIndices = CartesianOutputStoredIndices (hFile, boBinary, fMultiplier, nOffset, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    nIndices = SphericalOutputStoredIndices (hFile, boBinary, fMultiplier, nOffset, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    nIndices = CurveOutputStoredIndices (hFile, boBinary, fMultiplier, nOffset, psFuncData);
  	break;
  default:
  	// No vertices
    nIndices = 0;
    break;
  }

	return nIndices;
}

int CountStoredVertices (double fMultiplier, FuncPersist const * psFuncData) {
	int nVertices;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    nVertices = CartesianCountStoredVertices (fMultiplier, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    nVertices = SphericalCountStoredVertices (fMultiplier, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    nVertices = CurveCountStoredVertices (fMultiplier, psFuncData);
  	break;
  default:
  	// No vertices
    nVertices = 0;
    break;
  }

	return nVertices;
}

int CountStoredFaces (double fMultiplier, FuncPersist const * psFuncData) {
	int nFaces;

  switch (psFuncData->eType) {
  case FUNCTYPE_CARTESIAN:
    nFaces = CartesianCountStoredFaces (fMultiplier, psFuncData);
    break;
  case FUNCTYPE_SPHERICAL:
    nFaces = SphericalCountStoredFaces (fMultiplier, psFuncData);
    break;
  case FUNCTYPE_CURVE:
    nFaces = CurveCountStoredFaces (fMultiplier, psFuncData);
  	break;
  default:
  	// No vertices
    nFaces = 0;
    break;
  }

	return nFaces;
}


